require 'sdl'
require 'text'
require 'uconv'
require 'fpstimer'
require 'conf'
require 'ranking'

class Typing
  def initialize(conf)
	srand Time.now.to_i

	generater = nil
	if conf.type == "text" then
	  require 'markov_generater'
	  generater = MarkovGenerater.new(conf.filename)
	elsif conf.type == "dat" then
	  require 'flat_generater'
	  generater = FlatGenerater.new(conf.filename)
	else
	  conf.help
	end

	@words = Array.new
	for i in 0...1024
	  @words.push(Uconv.euctou8(generater.generate))
	end
	@word = 0

	SDL.init(SDL::INIT_VIDEO)
	SDL::TTF.init
	SDL::Event2.enableUNICODE

	@fonts = Array.new(20)
	for i in 0...20
	  @fonts[i] = SDL::TTF.open(conf.font, 16+1*i)
	end
	@fontmini = SDL::TTF.open(conf.font, 12)
	@dict = SDL::SKK::Dictionary.new
	@dict.load('jisyo', false)
	@table = SDL::SKK::RomKanaRuleTable.new('rule_table')

	@screen = SDL::setVideoMode(400, 400, 16, SDL::SWSURFACE)
	SDL::WM.setCaption("rssmt", "ruby sdl skk markov typing")

	@black = SDL::Surface.new(SDL::SWSURFACE, 400, 400, @screen)
	@black.setAlpha(SDL::SRCALPHA|SDL::RLEACCEL, 35)

	@scoreStr = Uconv.euctou8('Score')
	@levelStr = Uconv.euctou8('Level')
	@remainStr = Uconv.euctou8('Remain')

	@firstLevel = conf.level
	@smoke = conf.smoke
  end

  def init
	@context = SDL::SKK::Context.new(@dict, @table)
	@texts = Array.new
	@score = 0
	@timer = FPSTimer.new(20, 4)
	@remain = 10
	@level = @firstLevel
  end

  def title
	image = SDL::Surface.load('title.png')
	image.setColorKey(SDL::SRCCOLORKEY|SDL::RLEACCEL, 0)

	anyInput = false
	timer = FPSTimer.new(20, 20)
	while !anyInput
	  while event = SDL::Event2.poll do
		case event
		when SDL::Event2::Quit
		  exit
		when SDL::Event2::KeyDown
		  anyInput = true
		end
	  end

	  timer.wait

	  clear
	  SDL.blitSurface(image, 0, 0, 0, 0, @screen, rand(10), rand(10))

	  @screen.flip
	end
  end

  def run()
	loop do
	  @level += 0.002

	  if (rand(2000) < 13 - @texts.size * 4 || @texts.empty?) then
		@word = (@word+1) & 1023
		level = @level + 3.0
		@texts.push(Text.new(@words[@word], (rand(level)+level)*0.01))
	  end

	  @texts.each { |t| t.move }

	  num = @texts.size
	  @texts.delete_if { |t| t.z >= 20 }
	  num -= @texts.size

	  if num > 0 then
		@screen.fillRect(0, 0, 400, 400, @screen.mapRGB(255, 0, 0))
	  else
		clear
	  end

	  @remain -= num
	  if @remain < 0 then
		@remain = 0
		break
	  end

	  if !@timer.shouldSkip then
		@texts.each { |t| drawText t }
	  end

	  handleEvent()

	  if !@timer.shouldSkip then
		showInfo()
	  end

	  if !@timer.shouldSkip then
		@screen.flip
	  end

	  @timer.wait
	end

	scoreEntry
  end

  def clear
	if @smoke then
	  SDL.blitSurface(@black, 0, 0, 0, 0, @screen, 0, 0)
	else
	  @screen.fillRect(0, 0, 400, 400, 0)
	end
  end

  def showInfo
	scoreStr = @scoreStr + @score.to_i.to_s
	s = @fonts[1].textSize(scoreStr)
	@fonts[1].drawSolidUTF8(@screen, scoreStr, 400-s[0], 0, 255, 255, 255)

	levelStr = @levelStr + @level.to_i.to_s
	@fonts[1].drawSolidUTF8(@screen, levelStr, 0, 0, 255, 255, 255)

	remainStr = @remainStr + @remain.to_i.to_s
	s = @fonts[1].textSize(remainStr)
	@fonts[1].drawSolidUTF8(@screen, remainStr, 400-s[0], 382, 255, 255, 255)

  end

  def handleEvent
	while event = SDL::Event2.poll do
	  case event
	  when SDL::Event2::Quit
		exit
	  when SDL::Event2::KeyDown
		if event.sym == SDL::Key::ESCAPE then
		    # ۤˤäȥޥˡ̵
		    @context = SDL::SKK::Context.new(@dict, @table)
		elsif event.sym == SDL::Key::RETURN then
		  erasing()
		end

		@context.input(event)
	  end
	end

	if !@timer.shouldSkip then
	  font = @fonts[1]
	  if @context.str.size > 34 then
		font = @fontmini
	  end
	  sz = font.textSize(' ')
	  text_surface = @context.render_str(font, 255, 255, 255)
	  SDL.blitSurface(text_surface, 0, 0, 0, 0, @screen, 0, 400-sz[1] )
	end
  end

  def erasing()
	now = Uconv.euctou8(@context.str)
	@texts.each_index{
	  |i|
	  if @texts[i].text == now then
		@score += @texts[i].text.size * @texts[i].sz * (50 - @texts[i].z)

		# ۤˤäȥޥˡ̵
		@context = SDL::SKK::Context.new(@dict, @table)
		# ƱʸϤۤȤɤǤʤͤ
		@texts.delete_at(i)
	  end
	}
  end

  def drawText(text)
	if text.z >= 16 then
	  if @smoke then
		text.sprite = (text.sprite + 1) & 7
	  else
		text.sprite = (text.sprite + 1) & 1
	  end
	end

	if text.sprite != 0 then
	  return
	end

    font = @fonts[text.z.to_i]
	s = font.textSize(text.text)
	image = SDL::Surface.new(SDL::SWSURFACE, s[0], s[1], @screen)
	if text.white then
	  font.drawSolidUTF8(image, text.text, 0, 0, 255, 255, 255)
	else
	  font.drawSolidUTF8(image, text.text, 0, 0, 255, text.green, text.blue)
	end
	image.setColorKey(SDL::SRCCOLORKEY|SDL::RLEACCEL, 0)
	SDL.transformBlit(image, @screen, text.r, 1, 1, s[0]*0.5, s[1]*0.5,
					  text.x, text.y, 0)
  end

  def scoreEntry
	scoreStr = @scoreStr + @score.to_i.to_s
	s = @fonts[1].textSize(scoreStr)
	scoreText = Text.new(scoreStr, 1)
	scoreText.set(400-s[0]*0.5, s[1]*0.5, 1, -7, 6, 0.3, -9)
	scoreText.setWhite

	levelStr = @levelStr + @level.to_i.to_s
	s = @fonts[1].textSize(levelStr)
	levelText = Text.new(levelStr, 1)
	levelText.set(s[0]*0.5, s[1]*0.5, 1, 7, 6, 0.3, 9)
	levelText.setWhite

	timer = FPSTimer.new(20, 20)
	for i in 0...20
	  clear
	  timer.wait
	  @screen.flip
	end

	for i in 0...40
	  clear
	  scoreText.move
	  levelText.move
	  drawText(scoreText)
	  drawText(levelText)
	  timer.wait
	  @screen.flip
	end

	scoreText.setSpeed(0, -3, 0, 0)
	levelText.setSpeed(0, -3, 0, 0)

	for i in 0...40
	  clear
	  scoreText.move
	  levelText.move
	  drawText(scoreText)
	  drawText(levelText)
	  timer.wait
	  @screen.flip
	end

	ranking = Ranking.new
	t =
	  if ranking.isRankIn?(@score) then
		"You ranked in! Input your name."
	  else
		"Out of rank. Please try again."
	  end

	s = @fonts[4].textSize(t)
	for i in 0...t.size+1
	  clear
	  drawText(scoreText)
	  drawText(levelText)
	  sz = (i != 0) ? @fonts[4].textSize(t[0...i])[0] : 0;
	  @fonts[4].drawSolidUTF8(@screen, t[i,i+2], 200-s[0]/2 + sz,
							  200, 255, 255, 255)
	  timer.wait
	  @screen.flip
	end

	lpend = false

	if ranking.isRankIn?(@score) then

	  throwEventAway

	  while !lpend
		clear
		drawText(scoreText)
		drawText(levelText)
		s = @fonts[4].textSize(t)
		@fonts[4].drawSolidUTF8(@screen, t[0,t.size+1],
								200-s[0]/2, 200, 255, 255, 255)

		while event = SDL::Event2.poll do
		  case event
		  when SDL::Event2::Quit
			exit
		  when SDL::Event2::KeyDown
			if event.sym == SDL::Key::ESCAPE then
			  # ۤˤäȥޥˡ̵
			  @context = SDL::SKK::Context.new(@dict, @table)
			elsif event.sym == SDL::Key::RETURN then
			  if !@context.str.empty? then
				lpend = true
			  end
			end

			@context.input(event)
		  end
		end

		sz = @fonts[1].textSize(' ')
		text_surface = @context.render_str(@fonts[1], 255, 255, 255)
		SDL.blitSurface(text_surface, 0, 0, 0, 0, @screen, 0, 400-sz[1] )

		timer.wait
		@screen.flip
	  end

	  ranking.insert(@context.str, @score)
	  ranking.save

	  @screen.fillRect(0, 180, 400, 220, 0)
	  for i in 0...ranking.names.size
		if ranking.now != i then
		  @fonts[2].drawSolidUTF8(@screen, Uconv.euctou8(ranking.names[i]),
								  130, 200+i*18, 255, 255, 255)
		  @fonts[2].drawSolidUTF8(@screen, ranking.scores[i].to_s,
								  230, 200+i*18, 255, 255, 255)
		else
		  @fonts[2].drawSolidUTF8(@screen, Uconv.euctou8(ranking.names[i]),
								  130, 200+i*18, 255, 0, 0)
		  @fonts[2].drawSolidUTF8(@screen, ranking.scores[i].to_i.to_s,
								  230, 200+i*18, 255, 0, 0)
		end
	  end
	  @screen.flip
	end

	throwEventAway

	lpend = false
	while !lpend
	  while event = SDL::Event2.poll do
		case event
		when SDL::Event2::Quit
		  exit
		when SDL::Event2::KeyDown
		  lpend = true
		end
	  end
	end

  end

  def throwEventAway
	while event = SDL::Event2.poll do
	end
  end

end



